/*
 * Fooyin
 * Copyright © 2024, Luke Taylor <LukeT1@proton.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include <core/engine/audioloader.h>

#include <QAbstractListModel>

#include <unordered_map>

class QWidget;

namespace Fooyin {
class DecoderModel : public QAbstractListModel
{
    Q_OBJECT

public:
    enum Role
    {
        HasSettings = Qt::UserRole
    };

    using SettingsHandler    = std::function<void(QWidget*)>;
    using SettingsHandlerMap = std::unordered_map<QString, SettingsHandler>;

    using LoaderVariant = std::variant<std::vector<AudioLoader::LoaderEntry<DecoderCreator>>,
                                       std::vector<AudioLoader::LoaderEntry<ReaderCreator>>>;

    explicit DecoderModel(QObject* parent = nullptr);

    void setup(const LoaderVariant& loaders, SettingsHandlerMap settingsHandlers = {});
    [[nodiscard]] LoaderVariant loaders() const;
    bool showSettings(const QModelIndex& index, QWidget* parent);

    [[nodiscard]] Qt::ItemFlags flags(const QModelIndex& index) const override;
    [[nodiscard]] int rowCount(const QModelIndex& parent) const override;
    [[nodiscard]] QVariant data(const QModelIndex& index, int role) const override;
    [[nodiscard]] bool setData(const QModelIndex& index, const QVariant& value, int role) override;
    [[nodiscard]] QStringList mimeTypes() const override;
    [[nodiscard]] QMimeData* mimeData(const QModelIndexList& indexes) const override;
    [[nodiscard]] bool canDropMimeData(const QMimeData* data, Qt::DropAction action, int row, int column,
                                       const QModelIndex& parent) const override;
    [[nodiscard]] Qt::DropActions supportedDropActions() const override;
    [[nodiscard]] Qt::DropActions supportedDragActions() const override;
    bool dropMimeData(const QMimeData* data, Qt::DropAction action, int row, int column,
                      const QModelIndex& parent) override;

private:
    LoaderVariant m_loaders;
    SettingsHandlerMap m_settingsHandlers;
};
} // namespace Fooyin
