/****************************************************************************
** Meta object code from reading C++ file 'playlistuicontroller.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/playlist/playlistuicontroller.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playlistuicontroller.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlaylistUiController::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlaylistUiController",
        "showCurrentTrack",
        "",
        "selectTracks",
        "Fooyin::TrackIds",
        "ids",
        "filterTracks",
        "Fooyin::PlaylistTrackList",
        "tracks",
        "requestPlaylistFocus"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'showCurrentTrack'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'selectTracks'
        QtMocHelpers::SignalData<void(const Fooyin::TrackIds &)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'filterTracks'
        QtMocHelpers::SignalData<void(const Fooyin::PlaylistTrackList &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'requestPlaylistFocus'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlaylistUiController, qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlaylistUiController::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlaylistUiController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistUiController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->showCurrentTrack(); break;
        case 1: _t->selectTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackIds>>(_a[1]))); break;
        case 2: _t->filterTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrackList>>(_a[1]))); break;
        case 3: _t->requestPlaylistFocus(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 1:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackIds >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistUiController::*)()>(_a, &PlaylistUiController::showCurrentTrack, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistUiController::*)(const Fooyin::TrackIds & )>(_a, &PlaylistUiController::selectTracks, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistUiController::*)(const Fooyin::PlaylistTrackList & )>(_a, &PlaylistUiController::filterTracks, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistUiController::*)()>(_a, &PlaylistUiController::requestPlaylistFocus, 3))
            return;
    }
}

const QMetaObject *Fooyin::PlaylistUiController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlaylistUiController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20PlaylistUiControllerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::PlaylistUiController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlaylistUiController::showCurrentTrack()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void Fooyin::PlaylistUiController::selectTracks(const Fooyin::TrackIds & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::PlaylistUiController::filterTracks(const Fooyin::PlaylistTrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::PlaylistUiController::requestPlaylistFocus()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}
QT_WARNING_POP
